/*
 * cjbnet.c - CJB.NET update client for unix-like systems.
 *
 * (c) 2002 Kyle Donaldson <gile@livejournal.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <signal.h>
#include <syslog.h>

#define CJB_SERVER_PORT		80
#define CJB_SERVER_ADDR		"www.cjb.net"
#define CONFIG_FILE		"/etc/cjbnet.conf"
#define CJB_PRIO		(LOG_USER | LOG_WARNING)

static char format_str[] = "GET /cgi-bin/dynip.cgi?username=%s&password=%s HTTP/1.1\r\nHost: www.cjb.net\r\nConnection: close\r\nUser-Agent: cjbnet/1.2 (gilex.net/code/cjbnet.html)\r\n\r\n";
static int cjb_current_ipaddr = 0;

static struct __cjb_config {
	char * user;
	char * pass;
	char * interface;
	int interval;
} config[1];

int openhost (char * host, int prt)
{
	int sock;
	struct sockaddr_in addr;
	struct hostent * he;
	he = gethostbyname(host);
	if (he == NULL)
		goto _oh_error;
	sock = socket(AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto);
	if (sock == -1)
		goto _oh_error;
	memcpy(&addr.sin_addr, he->h_addr, he->h_length);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(prt);
	if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1)
		goto _oh_error;
	return sock;
_oh_error:
	syslog(CJB_PRIO, "Connect to %s failed: %m.", host);
	return -1;
}

char * cjb_dot_quad (u_int32_t addr)
{
	char * ret;
	int part[4];
	ret = malloc(16);
	if (!ret)
		return "0.0.0.0";
	part[0] = (addr >> 24);
	part[1] = ((addr >> 16) - ((addr >> 24) << 8));
	part[2] = ((addr >> 8) - ((addr >> 16) << 8));
	part[3] = ((addr) - ((addr >> 8) << 8));
	sprintf(ret, "%d.%d.%d.%d", part[0], part[1], part[2], part[3]);
	return ret;
}

int cjb_update_addr (u_int32_t ipaddr)
{
	int sock;
	int i;
	int ver = 0;
	char * msg;
	sock = openhost(CJB_SERVER_ADDR, CJB_SERVER_PORT);
	if (sock == -1)
		return -1;
	msg = malloc(strlen(format_str) + strlen(config->user) + strlen(config->pass) + 2);
	if (!msg) {
		syslog(CJB_PRIO, "Unable to allocate memory for HTTP message");
		return -1;
	}
	sprintf(msg, format_str, config->user, config->pass);
	write(sock, msg, strlen(msg));
	syslog(CJB_PRIO, "%s.cjb.net is now online.", config->user);
	free(msg);
	close(sock);
	syslog(CJB_PRIO, "IP address changed to %s.", cjb_dot_quad(ipaddr));
	return 0;
}

u_int32_t cjb_ipv4_addr (void)
{
	int sock;
	struct ifreq ifr;
	struct sockaddr_in * sdr;
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1)
		return 0;
	strcpy(ifr.ifr_name, config->interface);
	ifr.ifr_addr.sa_family = AF_INET;
	if (ioctl(sock, SIOCGIFADDR, &ifr) != 0)
		return 0;
	sdr = (struct sockaddr_in *) &ifr.ifr_addr;
	close(sock);
	return ntohl(sdr->sin_addr.s_addr);
}

int cjb_load_config (void)
{
	FILE * fd;
	char line[512];
	char v_name[100];
	char v_val[200];
	register int i;
	int lnc = 0;
	int dnc = 0;
	fd = fopen(CONFIG_FILE, "r");
	if (!fd) {
		fprintf(stderr, "Cannot open config file %s: %s.\n", CONFIG_FILE, strerror(errno));
		fprintf(stderr, "cjb.net cannot run without a configuration file, exiting...\n");
		return -1;
	}
	while (!feof(fd)) {
		lnc++;
		memset(line, 0, 512);
		memset(v_name, 0, 100);
		memset(v_val, 0, 200);
		fgets(line, 512, fd);
		if (strlen(line) < 5)
			continue;
		line[strlen(line) - 1] = '\0';
		for (i = 0; (isspace(line[i])) && (i < strlen(line)); i++);
		if (i > 0)
			strcpy(line, line + i);
		if (line[0] == '#')
			continue;
		for (i = 0; (!isspace(line[i])) && (i < strlen(line)); i++);
		strncpy(v_name, line, i);
		strcpy(line, line + i);
		for (i = 0; (isspace(line[i])) && (i < strlen(line)); i++);
		if (line[i] != '=') {
			fprintf(stderr, "%s:%d: Missing '=' separator.  Aborting...\n", CONFIG_FILE, lnc);
			fclose(fd);
			return -1;
		}
		for (i++; (isspace(line[i])) && (i < strlen(line)); i++);
		strcpy(v_val, line + i);
		if (!strcasecmp(v_name, "Username"))
			config->user = strdup(v_val);
		else if (!strcasecmp(v_name, "Password"))
			config->pass = strdup(v_val);
		else if (!strcasecmp(v_name, "Update-Interval"))
			config->interval = atoi(v_val) * 60;
		else if (!strcasecmp(v_name, "Network-Interface"))
			config->interface = strdup(v_val);
	}
	fclose(fd);
	return 0;
}

void cjb_signal (int sig)
{
	switch (sig) {
		case SIGHUP:
			syslog(CJB_PRIO, "reloading config file (active on next update)");
			cjb_load_config();
			break;
		default: break;
	}
}

int main (int argc, char ** argv)
{
	u_int32_t ipaddr;
	signal(SIGHUP, cjb_signal);
	openlog("cjb.net", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_USER);
	config->interval = 3600;
	if (cjb_load_config() != 0)
		exit(1);
	daemon(0, 0);
	syslog(CJB_PRIO, "cjbnet/1.2 now running on interface %s.", config->interface);
	while (1) {
		ipaddr = cjb_ipv4_addr();
		if (cjb_current_ipaddr != ipaddr) {
			if (cjb_update_addr(ipaddr) != 0)
				/*exit(1)*/;
			cjb_current_ipaddr = ipaddr;
		}
		sleep(config->interval);
	}
	closelog();
	exit(1);
}
